#include "detector.h"
#include "debug.h"
#include <stack>


using namespace std;


/// Transformation of OpenCV structure.
/// Just takes info from the IplImage structure and copies them to new TImage structure.
/// No memory copies are performed.
/// \param img The original IplImage
/// \returns Filled TImage structure
TImage image(IplImage * img)
{
    TImage tmp;
    tmp.width = img->width;
    tmp.height = img->height;
    tmp.widthStep = img->widthStep;
    tmp.imageData = img->imageData;
    return tmp;
}


Pyramid::Pyramid(CvSize base, int _levels_per_octave, int _octaves)
    : octaves(_octaves), levels_per_octave(_levels_per_octave)
{
    float f = pow(2.0, 1.0f/levels_per_octave); // scale factor
    CvSize _sz = base;
    for (int o = 0; o < octaves; ++o)
    {
        for (int i = 0; i < levels_per_octave; ++i)
        {
            int w = _sz.width / pow(f, i);
            int h = _sz.height / pow(f, i);
            IplImage * tmp = cvCreateImage(cvSize(w,h), IPL_DEPTH_8U, 1);
            cvZero(tmp);
            img.push_back(tmp);
        }
        _sz.width /= 2;
        _sz.height /= 2;
    }
}

void Pyramid::insert(IplImage * image)
{
    cvResize(image, img[0], CV_INTER_AREA);
    // first pass 
    for (unsigned i = levels_per_octave; i < img.size(); i += levels_per_octave)
    {
        cvPyrDown(img[i-levels_per_octave], img[i]);
    }
    // second pass
    for (int o = 0; o < octaves; ++o)
        for (int i = 1; i < levels_per_octave; ++i)
        {
            cvResize(img[o*levels_per_octave], img[o*levels_per_octave+i], CV_INTER_AREA);
        }
}



int abrDetectObjects(Pyramid & p, TClassifier * c, TRect * begin, TRect * end, int * hist)
{
    IplImage * base = p.img[0];
    int total = 0;
    vector<IplImage*>::iterator img;
    for (img = p.img.begin(); img != p.img.end(); ++img)
    {
        float scale = float(base->width) / (*img)->width;
        TImage _img = image(*img);
        //debug("img: %dx%d; ", _img.width, _img.height);
        setClassifierImageSize(c, _img.widthStep);
        int n = scanImage(&_img, c, begin+total, end, hist);
        //debug("n=%d; ", n);
        for (int i = total; i < total+n; ++i)
        {
            begin[i].x *= scale;
            begin[i].y *= scale;
            begin[i].width *= scale;
            begin[i].height *= scale;
        }
        total += n;
    }
    return total;
}


/// Calculate overlap of two rectangles (normalized to size of smaller rect).
/// \param a First rectangle
/// \param b Second rectangle
/// \returns Overlap area normalized by size of the smaller rect.
static inline float overlap(TRect * a, TRect * b)
{
    int x0 = max(a->x, b->x);
    int x1 = min(a->x+a->width, b->x+b->width);
    int y0 = max(a->y, b->y);
    int y1 = min(a->y+a->height, b->y+b->height);
    float area = min(a->width*a->height, b->width*b->height);
    if (x0 > x1 || y0 > y1) return 0;
    return (float)((x1-x0) * (y1-y0))/area;
}

/// Clustering of rectangles based on overlap.
/// \param src Source rectangles
/// \param dst Pre-allocated array for clustered rectangles
/// \param n Amount of 'src' rectangles
/// \param t Threshold for overlap
/// \returns Number of clustered rectangles.
int abrClusterDetections(TRect * src, TRect * dst, int n, float t, int minSz, float scale)
{
    float M[n*n]; // Overlap matric
    fill(M, M+n*n, 0.0f);
    
    int groups = 0;
    
    // Calc one to one overlap
    for (int i = 0; i < (n-1); ++i)
        for (int j = i+1; j < n; ++j)
        {
            M[i*n+j] = overlap(src+i, src+j);
        }

    bool mark[n]; // For marking of processed rectangles
    fill(mark, mark+n, true);

    stack<int> s; // Processing stack (just to avoid recursion)

    for (int i = 0; i < n; ++i)
    {
        if (mark[i]) // Marked item -> new group
        {
            int group[n]; // indices of rects in a group
            int c = 0; // Size of the group
            group[c++] = i; // Add i-th idx to the new group
            mark[i] = false; // Unmark - no further processing
            s.push(i);
            
            while (!s.empty()) // Add all overlapping rects and process them
            {
                int k = s.top();
                s.pop();
                for (int j = k+1; k < n; ++k)
                {
                    if (M[k*n+j] > t)
                    {
                        // Big overlap - add to the group and put to stack for processing
                        group[c++] = j;
                        s.push(j);
                        mark[j] = false;
                    }
                }
            }

            //if (c < minSz) continue;
            
            // Search for best rectangle in the group as a representative
            float best_response = -1E10f;
            int best_id = -1;
            for (int i = 0; i < c; ++i)
                if (src[group[i]].response > best_response)
                {
                    best_response = src[group[i]].response;
                    best_id = group[i];
                }

            dst[groups] = src[best_id];
            dst[groups].x *= scale;
            dst[groups].y *= scale;
            dst[groups].width *= scale;
            dst[groups].height *= scale;
            groups++;
        }
    }

    return groups;
}

