/*
 *   engine.cpp
 *   Classifier evaluation engine
 *   ----
 *   Department of Computer Graphics and Multimedia
 *   Faculty of Information Technology, Brno University of Technology
 *   ----
 */

#include "engine.h"
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>


/// Initialize 'TRect' structure.
/// New structure is created and filled (except for the 'response' field).
/// \param _x x coordinate
/// \param _y y coordinate
/// \param _w Width
/// \param _h Height
/// \returns New 'TRect' structure
TRect rect(int _x, int _y, int _w, int _h)
{
    TRect r = {_x, _y, _w, _h};
    return r;
}


void initClassifier(TClassifier * classifier)
{
    float * stgAlpha = classifier->alpha;
    unsigned s;
    for (s = 0; s < classifier->stageCount; ++s, stgAlpha += classifier->alphaCount)
    {
        switch (classifier->tp)
        {
        case LBP:
        case CSLBP:
        case LBP_22:
        case LRP:
        case LRD:
            {
                TStage * stage = (TStage*)classifier->stage + s;
                stage->alpha = stgAlpha;
                stage->offset = 0;
                break;
            }
        case HAAR:
            {
                THaarStage * stage = (THaarStage*)classifier->stage + s;
                stage->alpha = stgAlpha;
                stage->offset = 0;
                stage->range = stage->max - stage->min;
                stage->size = stage->w * stage->h;
                break;
            }
        default:
            assert("Unknown classifier");
            break;
        }
    }
}

void setClassifierImageSize(TClassifier * classifier, unsigned widthStep)
{
    unsigned s;
    for (s = 0; s < classifier->stageCount; ++s)
    {
        switch (classifier->tp)
        {
        case LBP:
        case CSLBP:
        case LBP_22:
        case LRP:
        case LRD:
            {
                TStage * stage = (TStage*)classifier->stage + s;
                stage->offset = stage->y * widthStep + stage->x;
                break;
            }
        case HAAR:
            {
                THaarStage * stage = (THaarStage*)classifier->stage + s;
                stage->offset = stage->y * widthStep + stage->x;
                break;
            }
        default:
            assert("Unknown classifier");
            break;
        }

    }
}


/// Sum of regions in 3x3 grid.
/// Sums values in regions and stores the results in a vector.
/// @param data Ptr to top-left corner 
static void sumRegions3x3(unsigned char * data, int w, int h, unsigned widthStep, int * v)
{
    unsigned blockStep = h * widthStep;
    widthStep -= w;

    // Prepare pointer array
    unsigned char * base[9] = {
        data, data+w, data+2*w,
        data+blockStep,data+blockStep+w,data+blockStep+2*w,
        data+2*blockStep,data+2*blockStep+w,data+2*blockStep+2*w,
    };

    int y;
    for (y = 0; y < h; ++y)
    {
        // go through all pixels in row and accumulate
        int x = 0;
        while (x < w)
        {
            int i;
            for (i = 0; i < 9; ++i)
            {
                v[i] += *base[i];
                ++base[i];
            }
            ++x;
        }
        // set pointers to next line 
        int i;
        for (i = 0; i < 9; ++i)
            base[i] += widthStep;
    }
}

static void sumRegions2x2(unsigned char * data, int w, int h, unsigned widthStep, int * v)
{
    unsigned blockStep = h * widthStep;
    widthStep -= w;

    // Prepare pointer array
    unsigned char * base[4] = {
        data, data+w,
        data+blockStep,data+blockStep+w
    };

    int y;
    for (y = 0; y < h; ++y)
    {
        // go through all pixels in row and accumulate
        int x = 0;
        while (x < w)
        {
            int i;
            for (i = 0; i < 4; ++i)
            {
                v[i] += *base[i];
                ++base[i];
            }
            ++x;
        }
        // set pointers to next line 
        int i;
        for (i = 0; i < 4; ++i)
            base[i] += widthStep;
    }
}
/// Calculate ranks of two elements of vector.
/// \param v Vector of values
/// \param n Number of values if the vector
/// \param a Index of the first value
/// \param b Index of the second value
/// \param rankA Rank of a-th item
/// \param rankB Rank of b-th item
static inline void calcRanks2(const int * v, const int n, const int a, const int b, int * rankA, int * rankB)
{
    *rankA = *rankB = 0;
    const int * data = v;
    while (data < v + 9)
    {
        if (v[a] > *data) ++(*rankA);
        if (v[b] > *data) ++(*rankB);
        ++data;
    }
}

/// Simple version of LRD evaluation.
/// \param image Input image
/// \param smpOffset Offset of the sample
/// \param s Stage stage parameters
/// \returns Response of the stage 's'
static float evalLRDStageSimple(TImage * image, unsigned smpOffset, int * r, void * s)
{
    TStage * stg = (TStage*)s;
    int values[9] = {0,0,0,0,0,0,0,0,0};
    // Get absolute address of feature in image
    unsigned char * base = (unsigned char*)(image->imageData + smpOffset + stg->offset);
    // Get sums in the feature blocks
    // There is no need to calculate the means because we care only about
    // rank of blocks not actual values 
    sumRegions3x3(base, stg->w, stg->h, image->widthStep, values);

    int countA = 0;
    int countB = 0;

    // calculate ranks
    calcRanks2(values, 9, stg->A, stg->B, &countA, &countB);

    int lrd = countA - countB + 8;

    *r = lrd;

    // return weak hypothesis response
    return stg->alpha[lrd];
}


/// Local Rank Pattern evaluation.
/// \param image Input image
/// \param smpOffset Offset of the sample
/// \param s Stage stage parameters
/// \returns Response of the stage 's'
static float evalLRPStageSimple(TImage * image, unsigned smpOffset, int * r, void * s)
{
    TStage * stg = (TStage*)s;
    int values[9] = {0,0,0,0,0,0,0,0,0};
    // Get absolute address of feature in image
    unsigned char * base = (unsigned char*)(image->imageData + smpOffset + stg->offset);
    // Get sums in the feature blocks
    // There is no need to calculate the means because we care only about
    // rank of blocks not actual values 
    sumRegions3x3(base, stg->w, stg->h, image->widthStep, values);

    int countA = 0;
    int countB = 0;

    calcRanks2(values, 9, stg->A, stg->B, &countA, &countB);

    int lrp = countA + 10 * countB;

    *r = lrp;

    // return weak hypothesis response
    return stg->alpha[lrp];
}


/// Order of samples in the feature grid
const int LBPOrder[8] = {0, 1, 2, 5, 8, 7, 6, 3};

/// Classic LBP evaluation.
/// Evaluation of 8 bit LBP code. Expects 9 samples in the input.
/// \param v Vector of samples
/// \returns 8bit LBP code
static inline int calcLBP(const int * v)
{
    int code = 0;
    int i;
    for (i = 0; i < 8; ++i)
    {
        code |= (v[LBPOrder[i]] > v[4]) << i;
        //code |= (abs(v[LBPOrder[i]] - v[4]) > 32) << i;
    }
    return code;
}

/// Center symmetric LBP (CS-LBP) evaluation.
/// Evaluation of 4bit CS-LBP code. Expects 9 samples in the input (v[4] /center/ is ignored)
/// \param v Vector of samples
/// \returns 4bit LBP code
static inline int calcCSLBP(const int * v)
{
    int code = 0;
    int i;
    for (i = 0; i < 4; ++i)
    {
        code |= (v[LBPOrder[i]] > v[LBPOrder[i+4]]) << i;
    }
    return code;
}


static inline int calcLBP_22(const int * v)
{
    int code = 0;
    code |= (v[0] > v[1]);
    code |= (v[2] > v[3]) << 1;
    code |= (v[0] > v[2]) << 2;
    code |= (v[1] > v[3]) << 3; 
    code |= (v[0] > v[3]) << 4;
    code |= (v[1] > v[2]) << 5;
    return code;
}


/// Local Binary Pattern evaluation.
/// \param image Input image
/// \param smpOffset Offset of the sample
/// \param s Stage stage parameters
/// \returns Response of the stage 's'
static float evalLBPStageSimple(TImage * image, unsigned smpOffset, int * r, void * s)
{
    TStage * stg = (TStage*)s;

    int values[9] = {0,0,0,0,0,0,0,0,0};
    // Get absolute address of feature in image
    unsigned char * base = (unsigned char*)(image->imageData + smpOffset + stg->offset);
    // Get sums in the feature blocks
    // There is no need to calculate the means because we care only about
    // rank of blocks not actual values 
    sumRegions3x3(base, stg->w, stg->h, image->widthStep, values);

    int lbp = calcLBP(values);

    *r = lbp;

    // return weak hypothesis response
    return stg->alpha[lbp];
}


/// Center-symmetric Local Binary Pattern evaluation.
/// \param image Input image
/// \param smpOffset Offset of the sample
/// \param s Stage stage parameters
/// \returns Response of the stage 's'
static float evalCSLBPStageSimple(TImage * image, unsigned smpOffset, int * r, void * s)
{
    TStage * stg = (TStage*)s;

    int values[9] = {0,0,0,0,0,0,0,0,0};
    // Get absolute address of feature in image
    unsigned char * base = (unsigned char*)(image->imageData + smpOffset + stg->offset);
    // Get sums in the feature blocks
    // There is no need to calculate the means because we care only about
    // rank of blocks not actual values 
    sumRegions3x3(base, stg->w, stg->h, image->widthStep, values);

    int lbp = calcCSLBP(values);

    *r = lbp;

    // return weak hypothesis response
    return stg->alpha[lbp];
}


/// 2x2 LBP eval
/// \param image Input image
/// \param smpOffset Offset of the sample
/// \param s Stage stage parameters
/// \returns Response of the stage 's'
static float evalLBP_22StageSimple(TImage * image, unsigned smpOffset, int * r, void * s)
{
    TStage * stg = (TStage*)s;

    int values[4] = {0,0,0,0};
    // Get absolute address of feature in image
    unsigned char * base = (unsigned char*)(image->imageData + smpOffset + stg->offset);
    // Get sums in the feature blocks
    // There is no need to calculate the means because we care only about
    // rank of blocks not actual values 
    sumRegions2x2(base, stg->w, stg->h, image->widthStep, values);

    int lbp = calcLBP_22(values);

    *r = lbp;

    // return weak hypothesis response
    return stg->alpha[lbp];

}
/// Minimum.
/// \param a First value
/// \param b Second value
/// \returns Minimum of the two values.
static inline int min(int a, int b)
{
    return (a < b) ? a : b;
}

/// Maximum.
/// \param a First value
/// \param b Second value
/// \returns Maximum of the two values.
static inline int max(int a, int b)
{
    return (a > b) ? a : b;
}

///////////
// Haar evaluation on the intensity image

/// Number of samples for different types of Haar features
static const int haarSampling[12]= {
    2, 1, // DH
    1, 2, // DV
    3, 1, // TH
    1, 3, // TV
    2, 2, // DIAG
    3, 3, // SRND
};

/// Scales for Haar normalization
static const int haarNormScale[6]= {
    1, // DH
    1, // DV
    2, // TH
    2, // TV
    2, // DIAG
    8, // SRND
};

/// Weights of samples
static const float haarWeights[6][16] = {
    {-1,  1},

    {-1,  1},

    {1, -2, 1},

    {1, -2, 1},

    {1, -1, -1, 1},

    {1, 1, 1, 1, -8, 1, 1, 1, 1},
};

// Normalization factor for Haar features
static float normFactor;

/// Sum values in an area.
/// \param base Top-left corner of the area (image is assumed to be 8bit, single channel)
/// \param w Width of the area
/// \param h Height of the area
/// \param widthStep Distance between two rows in bytes
/// \returns sum of the area
static inline int sumArea(unsigned char * base, int w, int h, int widthStep)
{
    int sum = 0;
    int y;
    for (y = 0; y < h; ++y, base += widthStep)
    {
        unsigned char * px = base;
        while (px < base + w)
        {
            sum += *px;
            ++px; 
        }
    }
    return sum;
}

/// Calculate sum and sum of squares of an area.
/// \param base Top-left corner of the area (image is assumed to be 8bit, single channel)
/// \param w Width of the area
/// \param h Height of the area
/// \param widthStep Distance between two rows in bytes
/// \param sum Sum of the area
/// \param sum2 Sum of squares of the area
static inline void sumArea2(unsigned char * base, int w, int h, int widthStep, unsigned * sum, unsigned * sum2)
{
    *sum = 0;
    *sum2 = 0;
    int y;
    for (y = 0; y < h; ++y, base += widthStep)
    {
        unsigned char * px = base;
        while (px < base + w)
        {
            *sum += *px;
            *sum2 += *px * *px;
            ++px;
        }
    }
}

/// Discretize a float value.
/// Discretizes a value to defined number of bins in defined range (lo .. lo+range).
/// Output values are in range '0'..'bins-1'.
/// \param f The value
/// \param lo Low limit of the range
/// \param range The range of values
/// \param bins Number of bins
/// \returns Bin to which the value falls. In the case the value is lower
/// than 'lo' it returns 0, if the value exceeds 'lo+range' it returns 'bins-1'
static inline int discretize(float f, float lo, float range, int bins)
{
    int bin = (int)((f - lo) / range * (bins-1) + 0.5);
    return min(bins, max(0, bin));
}


/// Evaluation of Haar feature.
/// Haar feature evaluation on the intensity image (i.e. slow as hell).
/// \param image Input image
/// \param smpOffset Offset of the sample image
/// \param s Feature parameters
static float evalHaarStageSimple1(TImage * image, unsigned smpOffset, int * r, void * s)
{
    THaarStage * stg = (THaarStage*)s;
    int xs = haarSampling[2*stg->tp];
    int ys = haarSampling[2*stg->tp+1];
    int norm = haarNormScale[stg->tp];
    const float * weights = haarWeights[stg->tp];
    unsigned char * base = (unsigned char*)(image->imageData + smpOffset + stg->offset);

    int x,y;
    float f = 0.0f;
    for (y = 0; y < ys; ++y, base+=stg->h*image->widthStep)
    {
        for (x = 0; x < xs; ++x)
        {
            float w = *weights++;
            unsigned sum = sumArea(base + x * stg->w, stg->w, stg->h, image->widthStep);
            f += (float)sum * w;
        }
    }

    // Normalize
    f = f / (norm * stg->size) / normFactor;

    // Discretize
    int bin = discretize(f, stg->min, stg->range, stg->bins);

    *r = bin;
    
    //fprintf(stderr, "x=%d,y=%d,w=%d,h=%d\n", stg->x, stg->y, stg->w, stg->h);
    //fprintf(stderr, "min=%f,max=%f,bins=%d\n", stg->min, stg->max,stg->bins);
    //fprintf(stderr, "f=%f,bin=%d\n", f, bin);

    // return hypothesis respone
    return stg->alpha[bin];

}

/////////////
typedef float (*StageEvalFunc)(TImage*, unsigned, int*, void*); ///< Stage evaluation function
typedef float (*ClassifierEvalFunc)(TImage*, unsigned, TClassifier*, float*, int*, int*, StageEvalFunc); ///< Classifier evaluation function


static float evalLRDClassifier(TImage * image, unsigned smpOffset, TClassifier * classifier, float * responses, int * f_responses, int * hypotheses, StageEvalFunc eval)
{

    float response = 0.0;
    // go through all stages and accumulate response
    TStage * stage;
    for (stage = (TStage*)classifier->stage; stage < (TStage*)classifier->stage + classifier->stageCount; ++stage)
    {
        *responses = eval(image, smpOffset, f_responses++, stage);
        response += *responses;
        responses++;
        //fprintf(stderr, "%f\n", *response);
        // Test the waldboost threshold
        if (response < stage->theta_b)
        {
            *hypotheses = (stage - (TStage*)classifier->stage) + 1;
            return response;
        }
    }
    *hypotheses = classifier->stageCount;
    // All stages have been evaluated - test final stage threshold
    return response;
}


static float evalHaarClassifier(TImage * image, unsigned smpOffset, TClassifier * classifier, float * responses, int * f_responses, int * hypotheses, StageEvalFunc eval)
{
    // get norm factor - standard deviation
    unsigned sum = 0;
    unsigned sum2 = 0;
    unsigned size = classifier->width*classifier->height;
    sumArea2((unsigned char*)image->imageData + smpOffset, classifier->width, classifier->height, image->widthStep, &sum, &sum2);
    normFactor = sqrt((float)(sum2) / size - pow((float)(sum) / size, 2));

    float response = 0.0;
    // go through all stages and accumulate response
    THaarStage * stage;
    for (stage = (THaarStage*)classifier->stage; stage < (THaarStage*)classifier->stage + classifier->stageCount; ++stage)
    {
        *responses = eval(image, smpOffset, f_responses++, stage);
        response += *responses;
        responses++;
        // Test the waldboost threshold
        if (response < stage->theta_b)
        {
            *hypotheses = (stage - (THaarStage*)classifier->stage) + 1;
            //fprintf(stderr, "\n\n");
            return response;
        }
    }
    *hypotheses = classifier->stageCount;
    // All stages have been evaluated - test final stage threshold
    return response;
}

/// Identification of functions for evaluation of particular classifier.
/// \param classifier The classifier
/// \param eval Stage evaluation function
/// \param ceval Classifier evaluation function
/// \returns Filled 'eval' and 'ceval'. 0 in either one means that the classifier naccot be evaluated.
static void getClassifierEvalFuncs(TClassifier * classifier, StageEvalFunc * eval, ClassifierEvalFunc * ceval)
{
    *eval = 0;
    *ceval = 0;

    switch (classifier->tp)
    {
        case LRD:
            *eval = evalLRDStageSimple;
            *ceval = evalLRDClassifier;
            break;
        case LRP:
            *eval = evalLRPStageSimple;
            *ceval = evalLRDClassifier;
            break;
        case LBP:
            *eval = evalLBPStageSimple;
            *ceval = evalLRDClassifier;
            break;
        case CSLBP:
            *eval = evalCSLBPStageSimple;
            *ceval = evalLRDClassifier;
            break;
        case LBP_22:
            *eval = evalLBP_22StageSimple;
            *ceval = evalLRDClassifier;
            break;
        case HAAR:
            *eval = evalHaarStageSimple1;
            *ceval = evalHaarClassifier;
        default:
            break;
    };

    assert(*eval != 0 && "Unsupported classifier type");
    assert(*ceval != 0 && "Unsupported classifier type");
}


float evalClassifier(TImage * image, TClassifier * classifier, unsigned offset, float * responses, int * f_responses, int * hypotheses)
{
    assert(image != 0);
    assert(classifier != 0);
    assert(offset < image->height * image->widthStep);
    assert(classifier->width <= image->width && classifier->height <= image->height);
    assert(responses != 0);
    assert(f_responses != 0);
    assert(hypotheses != 0);
    
    StageEvalFunc eval = 0;
    ClassifierEvalFunc ceval = 0;
    getClassifierEvalFuncs(classifier, &eval, &ceval);
    
    return ceval(image, offset, classifier, responses, f_responses, hypotheses, eval);
}


unsigned scanImage(TImage * image, TClassifier * classifier, TRect * results, TRect * end, int * hist)
{
    if (image->width < classifier->width || image->height < classifier->height)
        return 0;

    TRect * iter = results;
    if (iter >= end)
        return 0;
    
    StageEvalFunc eval = 0;
    ClassifierEvalFunc ceval = 0;
    getClassifierEvalFuncs(classifier, &eval, &ceval);
    // fprintf(stderr, "%d,%d\n", classifier->width, classifier->height);

    // Set pointer to position for first detection


    unsigned rowOffset = 0;
    unsigned y;
    float responses[classifier->stageCount];
    int f_responses[classifier->stageCount];
    int hypotheses = 0;
    
    for (y = 0; y < image->height-classifier->height; y+=1, rowOffset+=image->widthStep)
    {
        unsigned x;
        for (x = 0; x < image->width-classifier->width; x+=1)
        {
            // Evaluate classifier
            float response = ceval(image, rowOffset+x, classifier, responses, f_responses, &hypotheses, eval);
            if (hist) hist[hypotheses]++;
            if (hypotheses == classifier->stageCount && response > classifier->threshold) // write position of the detection
            {
                *iter = rect(x, y, classifier->width, classifier->height);
                iter->response = response;
                ++iter;
                if (iter >= end)
                    return iter - results;
            }
        }
    }

    return iter - results;
}


